#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from typing import List, Tuple, Dict, Set, Sequence, Union, Optional, Any
from sys import stderr

from defs import (
    HermesDecompiler,
    DecompiledFunctionBody,
    Environment,
    TokenString,
    ForInLoopInit,
    ForInLoopNextIter,
    RawToken,
    ReturnDirective,
    ThrowDirective,
    JumpNotCondition,
    JumpCondition,
    AssignmentToken,
    BasicBlock,
    LeftParenthesisToken,
    RightHandRegToken,
    RightParenthesisToken,
    LeftHandRegToken,
    NewInnerEnvironmentToken,
    NewEnvironmentToken,
    StoreToEnvironment,
    GetEnvironmentToken,
    LoadFromEnvironmentToken,
    FunctionTableIndex,
)

# Implementation for the "NameNonLocalClosureVariables" algorithm (see the .odt notes)

"""
    NOTES about the original Hermes `lib/VM/Interpreter.cpp` file
    handling instructions such as `GetEnvironment`, `CreateEnvironment`,
    `StoreToEnvironment`, `LoadFromEnvironment`:
        - "FRAME.getCalleeClosureUnsafe()->getEnvironment(runtime)"
            ^ This seems to fetch the operand 2 of Call* calls
            containing a closure reference for the current function
        - "curEnv->getParentEnvironment(runtime)"
            ^ This fetches the environement that was to the current
            function's passed-environment at the point of creating
            `curEnv` (through indirecting a linked list), or likely
            the caller's environment
"""


def pass4_name_closure_vars(
    state: HermesDecompiler, function_body: DecompiledFunctionBody
):

    AT = AssignmentToken

    TS = TokenString
    RT = RawToken
    LHRT = LeftHandRegToken
    RHRT = RightHandRegToken

    function_body.local_items: Dict[int, Environment] = {}

    parent_environment = function_body.parent_environment

    lines = function_body.statements
    for index, line in enumerate(lines):
        for token in line.tokens:
            if isinstance(token, NewEnvironmentToken):
                function_body.local_items[token.register] = Environment(
                    parent_environment,
                    (parent_environment.nesting_quantity + 1)
                    if parent_environment
                    else 0,
                    {},
                )
                line.tokens = []  # Silence this instruction in the produced decompiled code

            elif isinstance(token, NewInnerEnvironmentToken):
                outer_environment = function_body.local_items[token.parent_register]
                function_body.local_items[token.dest_register] = Environment(
                    outer_environment, (outer_environment.nesting_quantity + 1), {}
                )

            elif isinstance(token, GetEnvironmentToken):
                environment = parent_environment
                for nesting in range(token.nesting_level):
                    environment = environment.parent_environment

                function_body.local_items[token.register] = environment
                line.tokens = []  # Silence this instruction in the produced decompiled code

            elif isinstance(token, FunctionTableIndex):
                if token.environment_id is not None:
                    token.parent_environment = function_body.local_items[
                        token.environment_id
                    ]

            elif isinstance(token, StoreToEnvironment):
                varname = "_closure%d_slot%d" % (
                    function_body.local_items[token.env_register].nesting_quantity,
                    token.slot_index,
                )

                if (
                    token.slot_index
                    not in function_body.local_items[
                        token.env_register
                    ].slot_index_to_varname
                ):
                    function_body.local_items[token.env_register].slot_index_to_varname[
                        token.slot_index
                    ] = varname
                    line.tokens = [
                        RT("var " + varname),
                        AT(),
                        RHRT(token.value_register),
                    ]

                else:  # This a closure-referenced variable reassignment
                    line.tokens = [RT(varname), AT(), RHRT(token.value_register)]

            elif isinstance(token, LoadFromEnvironmentToken):
                var_name = "_closure%d_slot%d" % (
                    function_body.local_items[token.register].nesting_quantity,
                    token.slot_index,
                )

                line.tokens[2] = RT(var_name)



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from ctypes import (
    LittleEndianStructure,
    c_uint8,
    c_uint16,
    c_uint32,
    c_uint64,
    c_int8,
    c_int16,
    c_int32,
    c_int64,
)
from typing import Sequence, Union, Dict, List, Set
from io import BytesIO, BufferedReader
from os.path import dirname, realpath
from argparse import ArgumentParser
from enum import IntEnum, IntFlag
from hashlib import sha1
import sys

PARSERS_DIR = dirname(realpath(__file__))
sys.path.insert(0, PARSERS_DIR)

# The following imports are made from the current directory:
from hbc_bytecode_parser import parse_hbc_bytecode, get_parser, ParsedInstruction
from regexp_bytecode_parser import decompile_regex, parse_regex
from pretty_print import pretty_print_structure
from debug_info_parser import print_debug_info

"""

    Regarding the file signature to check for, please
    refer to:
    https://github.com/file/file/blob/FILE5_43/magic/Magdir/javascript#L19

!:mime application/javascript
# Hermes by Facebook https://hermesengine.dev/
# https://github.com/facebook/hermes/blob/master/include/hermes/\
# BCGen/HBC/BytecodeFileFormat.h#L24
0	lequad		0x1F1903C103BC1FC6	Hermes JavaScript bytecode
>8	lelong		x			\b, version %d
    +
    https://github.com/facebook/hermes/blob/main/include/hermes/BCGen/HBC/BytecodeFileFormat.h#L24
    
    Especially, find the header structure for the concerned
    file format here:
        https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/BCGen/HBC/BytecodeFileFormat.h#L71
        
    Parser in Python:
    
        https://github.com/bongtrop/hbctool/blob/v0.1.5/hbctool/hbc/__init__.py#L10
    
    [i] The function/method definitions in this file generally
        read from bottom to top.
"""

HEADER_MAGIC = int.from_bytes("Ἑρμῆ".encode("utf-16be"), "big")  # 0x1f1903c103bc1fc6

LATEST_BYTECODE_VERSION = 89

SHA1_NUM_BYTES = 20

# Find the latest bytecode version up to the current day here:
# https://github.com/facebook/hermes/blob/main/include/hermes/BCGen/HBC/BytecodeVersion.h#L23


class ProhibitInvoke(IntEnum):
    ProhibitCall = 0
    ProhibitConstruct = 1
    ProhibitNone = 2


class StringKind(IntEnum):
    String = 0
    Identifier = 1
    Predefined = 2  # Unused since version 0.3.0 - merged with Identifier


class HBCReader:
    # The following structure will be derived from the CTypesReader
    # class which is dynamically constructured below, with fields
    # that may vary depending upon the Hermes bytecode version.

    header: object
    function_headers: List[object]
    function_id_to_exc_handlers: Dict[int, List["HBCExceptionHandlerInfo"]]
    function_id_to_debug_offsets: Dict[int, "DebugOffsets"]
    string_kinds: List[StringKind]
    identifier_hashes: List[int]

    small_string_table: List[object]
    overflow_string_table: List[object]
    strings: List[str]

    parser_module: "module"
    file_buffer: BufferedReader

    # The following bytes can be sliced and decoded to "SLPArray"
    # objects using the "unpack_slp_array" function of
    # "serialized_literal_parser.py":
    arrays: bytes
    object_keys: bytes
    object_values: bytes

    bigint_values: List[int]

    regexp_table: List[object]
    regexp_storage: BytesIO

    cjs_modules: List[object]
    function_sources: List[object]

    debug_info_header: object
    debug_string_table: List[object]
    debug_string_storage: BytesIO
    debug_file_regions: List[object]
    sources_data_storage: BytesIO
    scope_desc_data_storage: BytesIO
    textified_data_storage: BytesIO
    string_table_storage: BytesIO

    def get_header_reader(
        self, bytecode_version: int = LATEST_BYTECODE_VERSION
    ) -> type:

        class CTypesReader(LittleEndianStructure):
            _pack_ = True

        fields = [
            ("magic", c_uint64),
            ("version", c_uint32),
            (
                "sourceHash",
                c_uint8 * SHA1_NUM_BYTES,
            ),  # This is a hash from the actual plain-text JS source.
            ("fileLength", c_uint32),  # Until the end of the BytecodeFileFooter.
            ("globalCodeIndex", c_uint32),
            ("functionCount", c_uint32),
            ("stringKindCount", c_uint32),  # Number of string kind entries.
            ("identifierCount", c_uint32),  # Number of strings which are identifiers.
            ("stringCount", c_uint32),  # Number of strings in the string table.
            (
                "overflowStringCount",
                c_uint32,
            ),  # Number of strings in the overflow table.
            ("stringStorageSize", c_uint32),  # Bytes in the blob of string contents.
        ]

        if bytecode_version >= 87:
            fields += [
                (
                    "bigIntCount",
                    c_uint32,
                ),  # Added in version 0.12.0 - Bytecode version 87 # number of bigints in the bigint table.
                (
                    "bigIntStorageSize",
                    c_uint32,
                ),  # Added in version 0.12.0 - Bytecode version 87 # Bytes in the bigint table.
            ]

        fields += [
            ("regExpCount", c_uint32),
            ("regExpStorageSize", c_uint32),
            ("arrayBufferSize", c_uint32),
            ("objKeyBufferSize", c_uint32),
            ("objValueBufferSize", c_uint32),
            (
                "cjsModuleOffset" if bytecode_version < 78 else "segmentID",
                c_uint32,
            ),  # Was called "cjsModuleOffset" before version 0.8.0 - Bytecode version 78 # The ID of this segment.
            ("cjsModuleCount", c_uint32),  # Number of modules.
        ]

        if bytecode_version >= 84:
            fields += [
                (
                    "functionSourceCount",
                    c_uint32,
                )  # Added in version 0.8.1 - Bytecode version 84 # Number of function sources preserved.
            ]

        fields += [
            ("debugInfoOffset", c_uint32),
            # Options (TODO: Are we decoding it correctly, in the right order?):
            ("staticBuiltins", c_uint8, 1),
            ("cjsModulesStaticallyResolved", c_uint8, 1),
            ("hasAsync", c_uint8, 1),
        ]

        # After these fields, padding bytes should align
        # the file cursor over a multiple of 32 bytes.

        CTypesReader._fields_ = fields

        return CTypesReader

    def get_small_func_header_reader(self) -> type:

        # The "SmallFunctionHeader" is stored just after the
        # "BytecodeFileHeader" in the .HBC Hermes bytecode file,
        # and each entry of it contains a field which will
        # indicate whether an "extended function header" will
        # be present for the concerned function near of the
        # end of the concerned file.

        # It is defined here in the Hermes C++ code base:
        # a) The fields themselves, in a generic macro:
        #  https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/BCGen/HBC/BytecodeFileFormat.h#L254
        # b) The structure declaration directive, instancying the aforementioned macro:
        # https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/BCGen/HBC/BytecodeFileFormat.h#L313
        # c) The FunctionHeaderFlag enum:
        # https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/BCGen/HBC/BytecodeFileFormat.h#L220

        # (the "FUNC_HEADER_FIELDS(CHECK_COPY_FIELD)" line is
        # another macro call which whether whether any of the
        # fields of the concerned structure - all the
        # corresponding fields being integers - overflow

        # In the hbctool project, it has been statically
        # defined here:
        # - (First) Bytecode version 59: https://github.com/bongtrop/hbctool/blob/v0.1.5/hbctool/hbc/hbc59/data/structure.json#L25
        # - (Last) Bytecode version 76: https://github.com/bongtrop/hbctool/blob/v0.1.5/hbctool/hbc/hbc76/data/structure.json#L25

        # - Does it has evolved through version of the .HBC format?
        # a) The field definition macro was the same within version 0.0.1 of Hermes. https://github.com/facebook/hermes/blob/v0.0.1/include/hermes/BCGen/HBC/BytecodeFileFormat.h#L225
        # b) The structure declaration directive neither: https://github.com/facebook/hermes/blob/v0.0.1/include/hermes/BCGen/HBC/BytecodeFileFormat.h#L284
        # c) The header flags bit field neither: https://github.com/facebook/hermes/blob/v0.0.1/include/hermes/BCGen/HBC/BytecodeFileFormat.h#L191
        # => We should conclude that the fields of the concerned structure should be stable across versions.

        class CTypesReader(LittleEndianStructure):
            _pack_ = True

            _fields_ = [
                # First word
                ("offset", c_uint32, 25),
                ("paramCount", c_uint32, 7),
                # Second word
                ("bytecodeSizeInBytes", c_uint32, 15),
                ("functionName", c_uint32, 17),
                # Third word
                ("infoOffset", c_uint32, 25),
                ("frameSize", c_uint32, 7),
                # Fourth word, with flags below
                ("environmentSize", c_uint8),
                ("highestReadCacheIndex", c_uint8),
                ("highestWriteCacheIndex", c_uint8),
                # Flags
                ("prohibitInvoke", c_uint8, 2),  # See enum: ProhibitInvoke
                ("strictMode", c_uint8, 1),
                ("hasExceptionHandler", c_uint8, 1),
                ("hasDebugInfo", c_uint8, 1),
                ("overflowed", c_uint8, 1),
                ("unused", c_uint8, 2),
            ]

        return CTypesReader

    def get_large_func_header_reader(self):

        class CTypesReader(LittleEndianStructure):
            _pack_ = True  # Packed?

            _fields_ = [
                # First word
                ("offset", c_uint32),
                ("paramCount", c_uint32),
                # Second word
                ("bytecodeSizeInBytes", c_uint32),
                ("functionName", c_uint32),
                # Third word
                ("infoOffset", c_uint32),
                ("frameSize", c_uint32),
                # Fourth word, with flags below
                ("environmentSize", c_uint32),
                ("highestReadCacheIndex", c_uint8),
                ("highestWriteCacheIndex", c_uint8),
                # Flags
                ("prohibitInvoke", c_uint8, 2),  # See enum: ProhibitInvoke
                ("strictMode", c_uint8, 1),
                ("hasExceptionHandler", c_uint8, 1),
                ("hasDebugInfo", c_uint8, 1),
                ("overflowed", c_uint8, 1),
                ("unused", c_uint8, 2),
            ]

        return CTypesReader

    def get_string_kind_entry_reader(self) -> type:

        # The layout of "StringKind::Entry" was changed with
        # bytecode version 71 (the release 0.3.0 of Hermes)

        # Cf. https://github.com/facebook/hermes/commit/5b4a501888ee9458615457d9bf0be05a4687e027

        class CTypesReader(LittleEndianStructure):
            _pack_ = True

        if self.header.version >= 71:
            fields = [("count", c_uint32, 31), ("kind", c_uint32, 1)]
        else:
            fields = [("count", c_uint32, 30), ("kind", c_uint32, 2)]

        CTypesReader._fields_ = fields

        return CTypesReader

    def get_small_string_table_entry_reader(self) -> type:
        # "struct SmallStringTableEntry" is defined here: https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/BCGen/HBC/BytecodeFileFormat.h#L169

        """
        Reminder of the wire level reading order of the C/C++
        bit fields:

        >>> x=X(); BytesIO(bytes([0b10111001, 0b01011010, 0b11101111, 0b10000011])).readinto(x); (x.isUTF16, x.isIdentifier, bin(x.offset), bin(x.length))
        4
        (1, 0, '0b1110111101011010101110', '0b10000011')
         A  B     CCCCCCCCDDDDDDDDEEEEEE      FFFFFFFF
        0b10111001, 0b01011010, 0b11101111, 0b10000011
          ABEEEEEE    DDDDDDDD    CCCCCCCC    FFFFFFFF

        This means that the "isIdentifier" field would
        occupy bit 7/23 of the "offset" field if the
        older format of this field was transposed into
        the newer corresponding format.
        """

        # The "isIdentifier" field was removed in version 0.0.3 -
        # bytecode version 56
        # Cf. https://github.com/facebook/hermes/commit/17816a14bad62dbccb2de75f62eb999b5973ad8b

        class CTypesReader(LittleEndianStructure):
            _pack_ = True

        if self.header.version >= 56:
            fields = [
                ("isUTF16", c_uint32, 1),
                ("offset", c_uint32, 23),
                ("length", c_uint32, 8),
            ]
        else:
            fields = [
                ("isUTF16", c_uint32, 1),
                (
                    "isIdentifier",
                    c_uint32,
                    1,
                ),  # Was removed because "Only used for asserts"
                ("offset", c_uint32, 22),
                ("length", c_uint32, 8),
            ]

        CTypesReader._fields_ = fields

        return CTypesReader

    def get_offset_length_pair_reader(self) -> type:

        # BigIntTableEntry is defines here as a structure:
        # https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/Support/BigIntSupport.h#L418

        class CTypesReader(LittleEndianStructure):
            _pack_ = True
            _fields_ = [("offset", c_uint32), ("length", c_uint32)]

        return CTypesReader

    def get_symbol_offset_pair_reader(self) -> type:

        class CTypesReader(LittleEndianStructure):
            _pack_ = True
            _fields_ = [("symbol_id", c_uint32), ("offset", c_uint32)]

        return CTypesReader

    def get_function_source_entry_reader(self) -> type:

        class CTypesReader(LittleEndianStructure):
            _pack_ = True
            _fields_ = [("function_id", c_uint32), ("string_id", c_uint32)]

        return CTypesReader

    def get_debug_info_header_reader(self) -> type:

        class CTypesReader(LittleEndianStructure):
            _pack_ = True

        if self.header.version >= 91:
            fields = [
                ("filename_count", c_uint32),
                ("filename_storage_size", c_uint32),
                ("file_region_count", c_uint32),
                (
                    "scope_desc_data_offset",
                    c_uint32,
                ),  # lexical_data_offset before up to version 91 and in version 93
                ("textified_data_offset", c_uint32),
                ("string_table_offset", c_uint32),
                ("debug_data_size", c_uint32),
            ]

        else:
            fields = [
                ("filename_count", c_uint32),
                ("filename_storage_size", c_uint32),
                ("file_region_count", c_uint32),
                ("scope_desc_data_offset", c_uint32),
                ("debug_data_size", c_uint32),
            ]

        CTypesReader._fields_ = fields

        return CTypesReader

    def get_debug_file_region_reader(self) -> type:

        class CTypesReader(LittleEndianStructure):
            _pack_ = True
            _fields_ = [
                ("from_address", c_uint32),
                ("filename_id", c_uint32),
                ("source_mapping_id", c_uint32),
            ]

        return CTypesReader

    def get_exception_handler_info_reader(self) -> type:

        class CTypesReader(LittleEndianStructure):
            _pack_ = True
            _fields_ = [("start", c_uint32), ("end", c_uint32), ("target", c_uint32)]

        return CTypesReader

    def get_debug_offsets_reader(self) -> type:

        class CTypesReader(LittleEndianStructure):
            _pack_ = True

        if self.header.version >= 91:
            fields = [
                ("source_locations", c_uint32),
                (
                    "scope_desc_data",
                    c_uint32,
                ),  # lexical_data before up to version 91 and in version 93
                ("textified_callees", c_uint32),
            ]

        else:
            fields = [("source_locations", c_uint32), ("scope_desc_data", c_uint32)]

        CTypesReader._fields_ = fields

        return CTypesReader

    def align_over_padding(self, padding_amount=4):

        # Each bytecode segment should be padded over
        # the size of an uint32_t

        current_pos = self.file_buffer.tell()

        if current_pos % padding_amount != 0:
            self.file_buffer.read(-current_pos % padding_amount)

    def read_header_from_buffer(self):

        # Check for a valid file header, and read the bytecode
        # format version (it will be required for properly
        # decoding the header fields)

        self.file_buffer.seek(0)
        magic = int.from_bytes(self.file_buffer.read(8), "little")
        if magic != HEADER_MAGIC:
            raise ValueError(
                "This file does not have the magic header for a Hermes bytecode file."
            )

        version = int.from_bytes(self.file_buffer.read(4), "little")
        self.parser_module = get_parser(version)

        self.file_buffer.read(SHA1_NUM_BYTES)  # Skip file hash

        file_size = int.from_bytes(self.file_buffer.read(4), "little")

        # Check the SHA-1 footer located at the end of the .HBC
        # bytecode file (the single-field BytecodeFileFooter structure)

        if version >= 75:
            self.file_buffer.seek(0)
            file_data = self.file_buffer.read(file_size)
            assert (
                sha1(file_data[:-SHA1_NUM_BYTES]).digest()
                == file_data[-SHA1_NUM_BYTES:]
            )

        # Decode the BytecodeFileHeader structure and
        # return it to the caller

        self.file_buffer.seek(0)

        header_reader = self.get_header_reader(version)()
        self.file_buffer.readinto(header_reader)

        self.align_over_padding(32)

        self.header = header_reader

    def read_functions(self):

        self.function_headers = []

        self.function_id_to_exc_handlers = {}
        self.function_id_to_debug_offsets = {}

        self.align_over_padding()

        reader = self.get_small_func_header_reader()
        reader_large = self.get_large_func_header_reader()

        for function_count in range(self.header.functionCount):
            function_header = reader()

            self.file_buffer.readinto(function_header)

            before_pos = self.file_buffer.tell()

            # Read the overflowed header, if any:
            if function_header.overflowed:
                new_offset = (function_header.infoOffset << 16) | function_header.offset
                function_header = reader_large()

                self.file_buffer.seek(new_offset)

                self.file_buffer.readinto(function_header)

            else:
                self.file_buffer.seek(function_header.infoOffset)

            self.function_headers.append(function_header)

            # Read the exception handler information, if any

            if function_header.hasExceptionHandler:
                self.align_over_padding()

                exc_headers_count = int.from_bytes(self.file_buffer.read(4), "little")
                exc_headers = (
                    self.get_exception_handler_info_reader() * exc_headers_count
                )()
                self.file_buffer.readinto(exc_headers)

                self.function_id_to_exc_handlers[function_count] = exc_headers

            # Read the debug information, if any

            if function_header.hasDebugInfo:
                self.align_over_padding()

                debug_header = self.get_debug_offsets_reader()()
                self.file_buffer.readinto(debug_header)

                self.function_id_to_debug_offsets[function_count] = debug_header

            self.file_buffer.seek(before_pos)

    def read_string_kinds(self):

        self.string_kinds = []

        self.align_over_padding()

        reader = self.get_string_kind_entry_reader()

        for string_kind_count in range(self.header.stringKindCount):
            string_kind = reader()

            self.file_buffer.readinto(string_kind)

            # Decode the run-length encoding instead of
            # storing the raw structure:
            self.string_kinds += [StringKind(string_kind.kind)] * string_kind.count

    def read_identifier_hashes(self):

        self.align_over_padding()

        self.identifier_hashes = [
            int.from_bytes(self.file_buffer.read(4), "little")
            for identifier_count in range(self.header.identifierCount)
        ]

    def read_small_string_table(self):

        self.align_over_padding()

        self.small_string_table = (
            self.get_small_string_table_entry_reader() * self.header.stringCount
        )()

        self.file_buffer.readinto(self.small_string_table)

    def read_overflow_string_table(self):

        self.align_over_padding()

        self.overflow_string_table = (
            self.get_offset_length_pair_reader() * self.header.overflowStringCount
        )()

        self.file_buffer.readinto(self.overflow_string_table)

    def read_string_storage(self):

        self.strings = []

        self.align_over_padding()

        string_storage = BytesIO(self.file_buffer.read(self.header.stringStorageSize))

        for string_count in range(self.header.stringCount):
            info = self.small_string_table[string_count]
            is_utf_16 = info.isUTF16
            if info.length == 0xFF:
                info = self.overflow_string_table[info.offset]
            length = info.length
            offset = info.offset

            if is_utf_16:
                length *= 2
            string_storage.seek(offset)
            string = string_storage.read(length)
            assert len(string) == length
            if is_utf_16:
                string = string.decode("utf-16", errors="surrogatepass")
            else:
                string = "".join(chr(char) for char in string)

            self.strings.append(string)

    def read_arrays(self):

        self.align_over_padding()
        self.arrays = self.file_buffer.read(self.header.arrayBufferSize)

        self.align_over_padding()
        self.object_keys = self.file_buffer.read(self.header.objKeyBufferSize)

        self.align_over_padding()
        self.object_values = self.file_buffer.read(self.header.objValueBufferSize)

    def read_bigints(self):

        self.align_over_padding()

        bigint_table = (
            self.get_offset_length_pair_reader() * self.header.bigIntCount
        )()

        self.file_buffer.readinto(bigint_table)

        self.align_over_padding()

        bigint_data = self.file_buffer.read(self.header.bigIntStorageSize)

        self.bigint_values = [
            int.from_bytes(
                bigint_data[entry.offset : entry.offset + entry.length], "little"
            )
            for entry in bigint_table
        ]

    def read_regexp(self):

        self.align_over_padding()

        self.regexp_table = (
            self.get_offset_length_pair_reader() * self.header.regExpCount
        )()

        self.file_buffer.readinto(self.regexp_table)

        self.align_over_padding()

        self.regexp_storage = BytesIO(
            self.file_buffer.read(self.header.regExpStorageSize)
        )

        # (We have implemented a RegExp bytecode decoder in the
        # neighboring "regexp_bytecode_parser.py" file.)

    def read_cjs_modules(self):

        self.align_over_padding()

        if self.header.cjsModulesStaticallyResolved and self.header.version < 77:
            self.cjs_modules = [
                int.from_bytes(self.file_buffer.read(4), "little")
                for module_count in range(self.header.cjsModuleCount)
            ]
        else:
            self.cjs_modules = (
                self.get_symbol_offset_pair_reader() * self.header.cjsModuleCount
            )()

            self.file_buffer.readinto(self.cjs_modules)

    def read_function_sources(self):

        self.align_over_padding()

        self.function_sources = (
            self.get_function_source_entry_reader() * self.header.functionSourceCount
        )()

        self.file_buffer.readinto(self.function_sources)

    def read_debug_info(self):

        self.file_buffer.seek(self.header.debugInfoOffset)

        self.debug_info_header = self.get_debug_info_header_reader()()
        self.file_buffer.readinto(self.debug_info_header)

        reader = self.get_offset_length_pair_reader()
        self.debug_string_table = (reader * self.debug_info_header.filename_count)()
        self.file_buffer.readinto(self.debug_string_table)

        self.debug_string_storage = BytesIO(
            self.file_buffer.read(self.debug_info_header.filename_storage_size)
        )

        reader = self.get_debug_file_region_reader()
        self.debug_file_regions = (reader * self.debug_info_header.file_region_count)()
        self.file_buffer.readinto(self.debug_file_regions)

        if self.header.version < 91:
            sources_data_size = self.debug_info_header.scope_desc_data_offset
            scope_desc_data_size = (
                self.debug_info_header.debug_data_size
                - self.debug_info_header.scope_desc_data_offset
            )

            self.sources_data_storage = BytesIO(
                self.file_buffer.read(sources_data_size)
            )
            self.scope_desc_data_storage = BytesIO(
                self.file_buffer.read(scope_desc_data_size)
            )
        else:
            sources_data_size = self.debug_info_header.scope_desc_data_offset
            scope_desc_data_size = (
                self.debug_info_header.textified_data_offset
                - self.debug_info_header.scope_desc_data_offset
            )
            textified_data_size = (
                self.debug_info_header.string_table_offset
                - self.debug_info_header.textified_data_offset
            )
            string_table_size = (
                self.debug_info_header.debug_data_size
                - self.debug_info_header.string_table_offset
            )

            self.sources_data_storage = BytesIO(
                self.file_buffer.read(sources_data_size)
            )
            self.scope_desc_data_storage = BytesIO(
                self.file_buffer.read(scope_desc_data_size)
            )

            self.textified_data_storage = BytesIO(
                self.file_buffer.read(textified_data_size)
            )
            self.string_table_storage = BytesIO(
                self.file_buffer.read(string_table_size)
            )

        # TODO: Improve debug information parsing with recent versions
        # of Hermes (bytecode versions >= 91)?

    def read_whole_file(self, file_buffer: BufferedReader):

        # Note: the structure of the bytecode file was
        # very different before version 0.0.3, the
        # order of the segments was not the same and
        # the padding/alignment logic was different too:
        # https://github.com/facebook/hermes/blob/v0.0.2/lib/BCGen/HBC/BytecodeDataProvider.cpp#L140

        # This was pre-public release of Hermes (Hermes was
        # announced with version 0.1.0 in July 2019:
        # https://engineering.fb.com/2019/07/12/android/hermes/
        # and is the default engine for React Native applications
        # since September 2022, or version 0.12.0:
        # https://reactnative.dev/blog/2022/09/05/version-070#hermes-as-default-engine)

        self.file_buffer = file_buffer

        self.read_header_from_buffer()  # Defines self.header

        self.read_functions()  # Defines self.function_headers

        self.read_string_kinds()  # Defines self.string_kinds

        self.read_identifier_hashes()  # Defines self.identifier_hashes

        self.read_small_string_table()  # Defines self.small_string_table

        self.read_overflow_string_table()  # Defines self.overflow_string_table

        self.read_string_storage()  # Defines self.string_storage and self.strings

        self.read_arrays()  # Defines self.arrays, self.object_keys, self.object_values

        self.bigint_values = []
        if self.header.version >= 87:
            self.read_bigints()  # Defines self.bigint_table and self.bigint_storage

        self.read_regexp()  # Defines self.regexp_table and self.regexp_storage

        self.read_cjs_modules()  # Defines self.cjs_modules

        if self.header.version >= 84:
            self.read_function_sources()  # Defines self.function_sources

        self.read_debug_info()  # Defines self.debug_info_header, self.debug_string_table, ...

        pass


def main():

    args = ArgumentParser()

    args.add_argument("input_file")

    args = args.parse_args()

    with open(args.input_file, "rb") as file_descriptor:
        hbc_reader = HBCReader()

        hbc_reader.read_whole_file(file_descriptor)

        pretty_print_structure(hbc_reader.header)

        """
        for function_header in hbc_reader.function_headers:
            pretty_print_structure(function_header)
            
            # Safety checks:
            assert function_header.unused == 0 and function_header.paramCount < 100
        """

        """
        # Disabled, we'll print it with strings now:
        for string_kind in hbc_reader.string_kinds:
            pretty_print_structure(string_kind)
        
        print()
        for identifier_count, identifier_hash in enumerate(hbc_reader.identifier_hashes):
            print('Identifier hash #%d: 0x%08x' % (
                identifier_count,
                identifier_hash
            ))
        print()
        
        for string in hbc_reader.small_string_table:
            pretty_print_structure(string)
        
        for overflow_string in hbc_reader.overflow_string_table:
            pretty_print_structure(overflow_string)
        """

        print()
        for string_kind, string in zip(hbc_reader.string_kinds, hbc_reader.strings):
            print("=> %s: %s " % (string_kind, repr(string)))

        for function_count, function_header in enumerate(hbc_reader.function_headers):
            # pretty_print_structure(function_header)
            print(
                "=> [Function #%d %s of %d bytes]: %d params @ offset 0x%08x"
                % (
                    function_count,
                    hbc_reader.strings[function_header.functionName],
                    function_header.bytecodeSizeInBytes,
                    function_header.paramCount,
                    function_header.offset,
                )
            )

            # Safety checks:
            assert function_header.unused == 0 and function_header.paramCount < 100

        # Commented, huge (and outdated)

        """
        print()
        for (arr_type, array) in [
            ('Array', hbc_reader.arrays),
            ('Object key', hbc_reader.object_keys),
            ('Object value', hbc_reader.object_values)
        ]:
            
            print()
            for item_count, item in enumerate(array.items):
                if item.tag_type in (TagType.LongStringTag, TagType.ShortStringTag, TagType.ByteStringTag):
                    print('[i] %s #%d of %s: %r' % (arr_type, item_count, item.tag_type, [(
                        hbc_reader.string_kinds[value],
                        hbc_reader.strings[value]
                    ) for value in item.values]))
                else:
                    print('[i] %s #%d of %s: %r' % (arr_type, item_count, item.tag_type, item.values))
        """

        print()
        print("=> BigInts:")
        for bigint in hbc_reader.bigint_values:
            print("=> BigInt:", bigint)

        print()
        for regexp_count, regexp in enumerate(hbc_reader.regexp_table):
            hbc_reader.regexp_storage.seek(regexp.offset)
            regexp_data = hbc_reader.regexp_storage.read(regexp.length)
            print("=> Regexp #%d: %s" % (regexp_count, regexp_data.hex()))
            print(
                "  => Decompiled: ",
                decompile_regex(
                    parse_regex(hbc_reader.header.version, BytesIO(regexp_data))
                ),
            )

        print()
        for cjs_module_count, cjs_module in enumerate(hbc_reader.cjs_modules):
            if isinstance(cjs_module, int):
                print("=> CommonJS module #%d: %d" % (cjs_module_count, cjs_module))
            else:
                print(
                    "=> CommonJS module #%d: %s @ %08x"
                    % (cjs_module_count, cjs_module.symbol_id, cjs_module.offset)
                )

        if getattr(hbc_reader, "function_sources", None):
            for function_source_count, function_source in enumerate(
                hbc_reader.function_sources
            ):
                print(
                    "=> Function source #%d: functionId %d = string @ %08x"
                    % (
                        function_source_count,
                        function_source.function_id,
                        function_source.string_id,
                    )
                )

        print()
        print("=> Debug data:")
        pretty_print_structure(hbc_reader.debug_info_header)
        for item in hbc_reader.debug_string_table:
            pretty_print_structure(item)
        print(hbc_reader.debug_string_storage.getvalue())
        for item in hbc_reader.debug_file_regions:
            pretty_print_structure(item)

        hbc_reader.sources_data_storage.seek(0)
        print_debug_info(hbc_reader.sources_data_storage, hbc_reader.header.version)

        print("  => Sources data:", hbc_reader.sources_data_storage.getvalue().hex())
        print(
            "  => Scope descriptor raw data:",
            hbc_reader.scope_desc_data_storage.getvalue().hex(),
        )
        if hbc_reader.header.version >= 91:
            print(
                "  => Textified data:",
                hbc_reader.textified_data_storage.getvalue().hex(),
            )
            print(
                "  => Raw variables and callees data:",
                hbc_reader.string_table_storage.getvalue().hex(),
            )


if __name__ == "__main__":
    main()



#!/usr/bin/python3
# -*- encoding: Utf-8 -*-
from ctypes import (
    LittleEndianStructure,
    c_bool,
    c_char,
    c_wchar,
    c_uint8,
    c_uint16,
    c_uint32,
)
from typing import List, Set, Sequence, Tuple, Any, Union, Dict, Optional
from logging import log, getLogger, DEBUG
from dataclasses import dataclass
from enum import IntEnum, IntFlag
from io import BytesIO
from re import sub

from pretty_print import pretty_print_structure

getLogger().setLevel(DEBUG)


"""
    The RegExp bytecode format is a custom format
    defined here:
        https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/Regex/RegexBytecode.h
        https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/Regex/RegexOpcodes.def
    
    Serialization code is provided here:
        https://github.com/facebook/hermes/blob/v0.12.0/lib/Support/RegExpSerialization.cpp#L282
    
"""

"""
    Additions to the format to note:

    Introduced in bytecode version 66 (version 0.3.0):
    https://github.com/facebook/hermes/commit/201bab747c200c31503b65e6d91db6b5ccda585b
    https://github.com/facebook/hermes/commit/892bf2cf461bfcaea117826ef06a965b1f67e658
            
        Emit Unicode bracket nodes
            REOP(U16Bracket)

        Emit Unicode-flavored instructions
            REOP(U16MatchCharICase32)
            REOP(U16MatchChar32)
            REOP(U16MatchAnyButNewline)
    
    Introduced in bytecode version 60 (version 0.1.1):
    https://github.com/facebook/hermes/commit/87334592a86a119e3618f8c514bb60badf84f2f4

    Optimize matching multiple literal characters
        REOP(MatchNChar8)
        REOP(MatchNCharICase8)
    
    Introduced in bytecode version 74 (version 0.4.0): https://github.com/facebook/hermes/commit/968a809749ac5c727c0fcbdaaa5fa113592f282d

    Implement RegExp dotall flag.
        REOP(MatchAny)
        REOP(U16MatchAny)
"""


class SyntaxFlags(IntFlag):
    ICASE = 1 << 0
    GLOBAL = 1 << 1  # NOSUBS until version 0.3.0. GLOBAL was added in 0.6.0.
    MULTILINE = 1 << 2
    UNICODE = 1 << 3  # UNICODE, later shorteneted to UCODE was added in version 0.3.0.
    DOTALL = 1 << 4  # DOTALL was added in version 0.4.0.
    STICKY = 1 << 5  # STICKY was added in version 0.6.0.
    INDICES = 1 << 6  # INDICES was added in bytecode version 96.


class MatchConstraintSet(IntFlag):
    MatchConstraintNonASCII = 1 << 0  # Requirement: contain 1 char value > 127
    MatchConstraintAnchoredAtStart = 1 << 1  # Requirement: match at pos 0
    MatchConstraintNonEmpty = 1 << 2  # Requirement: contain at least 1 char


MIN_BC_VERSION = 51


class RegexBytecodeHeader(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("markedCount", c_uint16),
        ("loopCount", c_uint16),
        ("syntaxFlags", c_uint8),  # Enum: SyntaxFlags
        ("constraints", c_uint8),  # Enum: MatchConstraintSet
    ]


def get_opcodes_enum(bytecode_version: int):

    # From https://github.com/facebook/hermes/blob/v0.12.0/include/hermes/Regex/RegexOpcodes.def

    opcode_name_and_min_version: List[Tuple[str, int]] = [
        ("Goal", MIN_BC_VERSION),
        ("LeftAnchor", MIN_BC_VERSION),
        ("RightAnchor", MIN_BC_VERSION),
        ("MatchAny", 74),  # Version 0.4.0
        ("U16MatchAny", 74),  # Version 0.4.0
        ("MatchAnyButNewline", MIN_BC_VERSION),
        ("U16MatchAnyButNewline", 66),  # Version 0.3.0
        ("MatchChar8", MIN_BC_VERSION),
        ("MatchChar16", MIN_BC_VERSION),
        ("U16MatchChar32", 66),  # Version 0.3.0
        ("MatchNChar8", 60),  # Version 0.1.1
        ("MatchNCharICase8", 60),  # Version 0.1.1
        ("MatchCharICase8", MIN_BC_VERSION),
        ("MatchCharICase16", MIN_BC_VERSION),
        ("U16MatchCharICase32", 66),  # Version 0.3.0
        ("Alternation", MIN_BC_VERSION),
        ("Jump32", MIN_BC_VERSION),
        ("Bracket", MIN_BC_VERSION),
        ("U16Bracket", 66),  # Version 0.3.0
        ("BeginMarkedSubexpression", MIN_BC_VERSION),
        ("EndMarkedSubexpression", MIN_BC_VERSION),
        ("BackRef", MIN_BC_VERSION),
        ("WordBoundary", MIN_BC_VERSION),
        (
            "Lookaround" if bytecode_version >= 67 else "Lookahead",
            MIN_BC_VERSION,
        ),  # Lookahead before version 0.3.0, one less field
        ("BeginLoop" if bytecode_version >= 79 else "BeginLoopPre79", MIN_BC_VERSION),
        ("EndLoop", MIN_BC_VERSION),
        ("BeginSimpleLoop", MIN_BC_VERSION),
        ("EndSimpleLoop", MIN_BC_VERSION),
        ("Width1Loop", MIN_BC_VERSION),
    ]

    Opcodes = IntEnum(
        "Opcodes",
        [
            opcode_name
            for opcode_name, min_version in opcode_name_and_min_version
            if bytecode_version >= min_version
        ],
        start=0,
    )

    return Opcodes


# The following structures have originally been generated
# by the "../utils/bytecode_structs_parser.py"


class GoalInsn(LittleEndianStructure):
    _pack_ = True


class LeftAnchorInsn(LittleEndianStructure):
    _pack_ = True


class RightAnchorInsn(LittleEndianStructure):
    _pack_ = True


class MatchAnyInsn(LittleEndianStructure):
    _pack_ = True


class U16MatchAnyInsn(LittleEndianStructure):
    _pack_ = True


class MatchAnyButNewlineInsn(LittleEndianStructure):
    _pack_ = True


class U16MatchAnyButNewlineInsn(LittleEndianStructure):
    _pack_ = True


class MatchChar8Insn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("c", c_uint8)]


class MatchChar16Insn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("c", c_uint16)]


class U16MatchChar32Insn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("c", c_uint32)]


class MatchCharICase8Insn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("c", c_uint8)]


class MatchCharICase16Insn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("c", c_uint16)]


class U16MatchCharICase32Insn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("c", c_uint32)]


class AlternationInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("secondaryBranch", c_uint32),  # hermes::regex::JumpTarget32
        ("primaryConstraints", c_uint8),  # hermes::regex::MatchConstraintSet
        ("secondaryConstraints", c_uint8),  # hermes::regex::MatchConstraintSet
    ]


class Jump32Insn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("target", c_uint32)  # hermes::regex::JumpTarget32
    ]


class BackRefInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("mexp", c_uint16)]


class BracketRange32(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("start", c_uint32), ("end", c_uint32)]


class BracketRange16(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("start", c_uint16), ("end", c_uint16)]


class BracketInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("rangeCount", c_uint32),
        ("negate", c_uint8, 1),
        ("positiveCharClasses", c_uint8, 3),
        ("negativeCharClasses", c_uint8, 3),
    ]

    brackets: List[Union[BracketRange32, BracketRange16]]


class U16BracketInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("rangeCount", c_uint32),
        ("negate", c_uint8, 1),
        ("positiveCharClasses", c_uint8, 3),
        ("negativeCharClasses", c_uint8, 3),
    ]

    brackets: List[BracketRange32]


class MatchNChar8Insn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("charCount", c_uint8)]

    chars: bytes


class MatchNCharICase8Insn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("charCount", c_uint8)]

    chars: bytes


class WordBoundaryInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("invert", c_bool)]


class BeginMarkedSubexpressionInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("mexp", c_uint16)]


class EndMarkedSubexpressionInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [("mexp", c_uint16)]


class LookaroundInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("invert", c_bool),
        ("forwards", c_bool),
        ("constraints", c_uint8),  # hermes::regex::MatchConstraintSet
        ("mexpBegin", c_uint16),
        ("mexpEnd", c_uint16),
        ("continuation", c_uint32),  # hermes::regex::JumpTarget32
    ]


class LookaheadInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("invert", c_bool),
        ("constraints", c_uint8),  # hermes::regex::MatchConstraintSet
        ("mexpBegin", c_uint16),
        ("mexpEnd", c_uint16),
        ("continuation", c_uint32),  # hermes::regex::JumpTarget32
    ]


class BeginLoopInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("loopId", c_uint32),
        ("min", c_uint32),
        ("max", c_uint32),
        ("mexpBegin", c_uint16),
        ("mexpEnd", c_uint16),
        ("greedy", c_bool),
        ("loopeeConstraints", c_uint8),  # hermes::regex::MatchConstraintSet
        ("notTakenTarget", c_uint32),  # hermes::regex::JumpTarget32
    ]


class BeginLoopPre79Insn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("loopId", c_uint32),
        ("min", c_uint32),
        ("max", c_uint32),
        ("mexpBegin", c_uint32),
        ("mexpEnd", c_uint32),
        ("greedy", c_bool),
        ("loopeeConstraints", c_uint8),  # hermes::regex::MatchConstraintSet
        ("notTakenTarget", c_uint32),  # hermes::regex::JumpTarget32
    ]


class EndLoopInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("target", c_uint32)  # hermes::regex::JumpTarget32
    ]


class BeginSimpleLoopInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("loopeeConstraints", c_uint8),  # hermes::regex::MatchConstraintSet
        ("notTakenTarget", c_uint32),  # hermes::regex::JumpTarget32
    ]


class EndSimpleLoopInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("target", c_uint32)  # hermes::regex::JumpTarget32
    ]


class Width1LoopInsn(LittleEndianStructure):
    _pack_ = True
    _fields_ = [
        ("loopId", c_uint32),
        ("min", c_uint32),
        ("max", c_uint32),
        ("greedy", c_bool),
        ("notTakenTarget", c_uint32),  # hermes::regex::JumpTarget32
    ]


class ParsedRegex:
    header: RegexBytecodeHeader
    instructions: List[object]

    def __init__(self):
        self.header = None
        self.instructions = []


# These functions are used to escape substrings when decompiling
# Hermes-compiled regular expression.

# For a list of which caracters to escape in EcmaScript regular
# expressions, please see:
# https://262.ecma-international.org/5.1/#sec-15.10.1

REGULAR_ESCAPE_MAP = {
    # Removed: '&~# ' and non-printable + '\\' ranges which
    # will be escaped by the unicode_escape codec, added: '/
    char: "\\" + chr(char)
    for char in b"()[]{}?*+|^$/."
}
RANGE_ESCAPE_MAP = {char: "\\" + chr(char) for char in b"[]-"}


def escape(string: str):
    return string.encode("unicode_escape").decode("ascii").translate(REGULAR_ESCAPE_MAP)


def escape_range(string: str):
    return string.encode("unicode_escape").decode("ascii").translate(RANGE_ESCAPE_MAP)


def parse_regex(bytecode_version: int, input_bytes: BytesIO) -> ParsedRegex:

    header = RegexBytecodeHeader()
    input_bytes.readinto(header)

    output = ParsedRegex()
    output.header = header

    # log(DEBUG, '_') # DEBUG
    # log(DEBUG, '=' * 34) # DEBUG
    # log(DEBUG, 'Parsed header:') # DEBUG
    # pretty_print_structure(header) # DEBUG

    opcode_enum = get_opcodes_enum(bytecode_version)

    while True:
        original_pos = input_bytes.tell() - 6

        next_opcode = input_bytes.read(1)
        if not next_opcode:
            break
        opcode = next_opcode[0]

        op_structure = globals()[opcode_enum(opcode).name + "Insn"]()
        input_bytes.readinto(op_structure)

        op_structure.original_pos = original_pos

        # log(DEBUG, '=' * 34) # DEBUG
        # log(DEBUG, 'Parsed %r:' % op_structure) # DEBUG
        # pretty_print_structure(op_structure) # DEBUG

        if opcode in (opcode_enum.Bracket, opcode_enum.U16Bracket):
            op_structure.brackets = []

            for count in range(op_structure.rangeCount):
                # BracketRange16 was changed to BracketRange32 in version 0.3.0 -
                # https://github.com/facebook/hermes/commit/28436fbfe9f4ac6a064dd22bd46122117d1080b7
                if bytecode_version >= 65:
                    bracket = BracketRange32()
                else:
                    bracket = BracketRange16()
                input_bytes.readinto(bracket)
                op_structure.brackets.append(bracket)

        elif opcode in (opcode_enum.MatchNChar8, opcode_enum.MatchNCharICase8):
            op_structure.chars = input_bytes.read(op_structure.charCount)

        output.instructions.append(op_structure)

    return output


def loop_ending_to_string(instruction):
    if instruction.min == 0 and instruction.max == 0xFFFFFFFF:
        output = "*"
    elif instruction.min == 0 and instruction.max == 1:
        output = "?"
    elif instruction.min == 1 and instruction.max == 0xFFFFFFFF:
        output = "+"
    elif instruction.max == 0xFFFFFFFF:
        output = "{%d,}" % instruction.min
    elif instruction.min == instruction.max:
        output = "{%d}" % instruction.min
    else:
        output = "{%d,%d}" % (instruction.min, instruction.max)
    if not instruction.greedy and output in "*+":
        output += "?"
    return output


class CharacterClass(IntFlag):
    Digits = 1 << 0
    Spaces = 1 << 1
    Words = 1 << 2


def decompile_regex(regex: ParsedRegex):

    output: str = ""

    secondary_branch_to_alternation_pos: Dict[int, int] = {}
    alternation_pos_to_jump_target: Dict[int, int] = {}

    pending_1width_instructions: List[Width1LoopInsn] = []
    pending_lookaround_instructions: List[Union[LookaroundInst, LookaheadInst]] = []
    pending_beginloop_instructions: List[Union[BeginLoopInst, BeginLoopPre78Inst]] = []

    for instruction in regex.instructions:
        while instruction.original_pos in alternation_pos_to_jump_target.values():
            output += ")"  # Close alternations
            alternation_pos = next(
                (
                    key
                    for key, value in alternation_pos_to_jump_target.items()
                    if value == instruction.original_pos
                ),
                None,
            )
            del alternation_pos_to_jump_target[alternation_pos]

        if isinstance(instruction, BeginMarkedSubexpressionInsn):
            output += "("
        elif isinstance(instruction, EndMarkedSubexpressionInsn):
            output += ")"
        elif (
            isinstance(instruction, BeginLoopInsn)
            or isinstance(instruction, BeginLoopPre79Insn)
            or isinstance(instruction, BeginSimpleLoopInsn)
        ):
            output += "(?:"
            pending_beginloop_instructions.append(instruction)
        elif isinstance(instruction, Width1LoopInsn):
            pending_1width_instructions.append(instruction)
            continue
        elif isinstance(instruction, LookaroundInsn) or isinstance(
            instruction, LookaheadInsn
        ):
            output += "(?"
            if not getattr(instruction, "forwards", True):
                output += "<"
            output += "!" if instruction.invert else "="
            pending_lookaround_instructions.append(instruction)
        elif isinstance(instruction, BracketInsn):
            display_brackets = (
                instruction.brackets
                or instruction.negate
                or (
                    bin(instruction.positiveCharClasses).count("1")
                    + bin(instruction.negativeCharClasses).count("1")
                    > 1
                )
            )
            display_brackets = display_brackets and not (
                instruction.brackets
                and not instruction.positiveCharClasses
                and not instruction.negativeCharClasses
                and not instruction.negate
                and len(instruction.brackets) == 1
                and instruction.brackets[0].start == instruction.brackets[0].end
            )
            if display_brackets:
                output += "["
            if instruction.negate:
                output += "^"
            for bracket in instruction.brackets:
                if bracket.start == bracket.end:
                    output += escape_range(chr(bracket.start))
                elif bracket.start + 1 == bracket.end:
                    output += escape_range(chr(bracket.start) + chr(bracket.end))
                else:
                    output += (
                        escape_range(chr(bracket.start))
                        + "-"
                        + escape_range(chr(bracket.end))
                    )
            for flag, char in [
                (CharacterClass.Digits, r"\d"),
                (CharacterClass.Spaces, r"\s"),
                (CharacterClass.Words, r"\w"),
            ]:
                if instruction.positiveCharClasses & flag:
                    output += char

                if instruction.negativeCharClasses & flag:
                    output += char.upper()

            if display_brackets:
                output += "]"
        elif isinstance(instruction, WordBoundaryInsn):
            if instruction.invert:
                output += r"\B"
            else:
                output += r"\b"
        elif isinstance(instruction, BackRefInsn):
            output += "\\%d" % instruction.mexp
        elif (
            isinstance(instruction, MatchAnyInsn)
            or isinstance(instruction, MatchAnyButNewlineInsn)
            or isinstance(instruction, U16MatchAnyInsn)
            or isinstance(instruction, U16MatchAnyButNewlineInsn)
        ):
            output += "."
        elif isinstance(instruction, LeftAnchorInsn):
            output += "^"
        elif isinstance(instruction, RightAnchorInsn):
            output += "$"
        elif isinstance(instruction, GoalInsn):
            if pending_lookaround_instructions:
                output += ")"
                pending_lookaround_instructions.pop()
            else:
                pass  # End of RegExp
        elif (
            isinstance(instruction, MatchChar8Insn)
            or isinstance(instruction, MatchChar16Insn)
            or isinstance(instruction, U16MatchChar32Insn)
            or isinstance(instruction, MatchCharICase8Insn)
            or isinstance(instruction, MatchCharICase16Insn)
            or isinstance(instruction, U16MatchCharICase32Insn)
        ):
            output += escape(chr(instruction.c))
        elif isinstance(instruction, EndLoopInsn) or isinstance(
            instruction, EndSimpleLoopInsn
        ):
            output += ")"
            begin_loop_instruction = pending_beginloop_instructions.pop()
            if isinstance(begin_loop_instruction, BeginSimpleLoopInsn):
                output += "*"
            else:
                output += loop_ending_to_string(begin_loop_instruction)
        elif isinstance(instruction, AlternationInsn):
            if instruction.original_pos not in secondary_branch_to_alternation_pos:
                output += "(?:"
                secondary_branch_to_alternation_pos[instruction.secondaryBranch] = (
                    instruction.original_pos
                )
            else:
                secondary_branch_to_alternation_pos[instruction.secondaryBranch] = (
                    secondary_branch_to_alternation_pos[instruction.original_pos]
                )  # Chained alternation
        elif isinstance(instruction, Jump32Insn):
            output += "|"
            alternation_pos_to_jump_target[
                secondary_branch_to_alternation_pos[instruction.original_pos + 5]
            ] = instruction.target
        elif isinstance(instruction, MatchNChar8Insn) or isinstance(
            instruction, MatchNCharICase8Insn
        ):
            output += escape(instruction.chars.decode("utf-8"))
        # To check again maybe later: Lookaround, loop, backref, jump, word boundary.. . SVG RegexP Usage...
        # Check code conformance and v0.0.1-more recent code conformance...
        # Test CASE;
        # =
        else:
            raise NotImplementedError("Please disassemble opcode: %r" % instruction)

        while pending_1width_instructions:
            instruction: Width1LoopInsn = pending_1width_instructions.pop()

            output += loop_ending_to_string(instruction)

    output = "/%s/" % output
    for flag, char in [
        (SyntaxFlags.ICASE, "i"),
        (SyntaxFlags.GLOBAL, "g"),
        (SyntaxFlags.MULTILINE, "m"),
        (SyntaxFlags.UNICODE, "u"),
        (SyntaxFlags.DOTALL, "s"),
        (SyntaxFlags.STICKY, "y"),
        (SyntaxFlags.INDICES, "d"),
    ]:
        if regex.header.syntaxFlags & flag:
            output += char

    # TODO : Create a real "test" file from the Hermes test file,
    # and do it for the global .HBC bytecode file header as well...

    # For readibility, avoid nesting (?:) and ()
    # when a BeginMarkedSubexpression appears within
    # a BeginLoop instruction:
    output = sub(
        r"(?<!\\)(?:\(\?:\(|\(\(\?:)"  # Match for "(?:(" or "((?:" not preceded with a "\"
        + r"([^()]*?"  # Capture the in-between in "\1"
        + r"[^\\)])\)\)",  # Match for "))" not preceded with a "\"
        r"(\1)",  # Replace "(?:(" or "((?:" SOMETHING "))" with "(" SOMETHING ")"
        output,
    )

    return output


def format_struct_to_single_line(struct: object) -> str:

    output: str = ""
    if hasattr(struct, "original_pos"):
        output += "(Pos: %d)  " % struct.original_pos

    output += struct.__class__.__name__.replace("Insn", "", 1)

    if getattr(struct, "_fields_", None):
        for field in struct._fields_:
            if field[0] == "syntaxFlags":
                output += " {%s=%r}" % (field[0], SyntaxFlags(struct.syntaxFlags))
            elif "constraint" in field[0].lower():
                output += " {%s=%r}" % (
                    field[0],
                    MatchConstraintSet(getattr(struct, field[0])),
                )
            elif "charclasses" in field[0].lower():
                output += " {%s=%r}" % (
                    field[0],
                    CharacterClass(getattr(struct, field[0])),
                )
            elif field[0] == "c":
                output += " {c=%s}" % repr(chr(struct.c))
            else:
                output += " {%s=%s}" % (field[0], getattr(struct, field[0]))

    if getattr(struct, "brackets", None):
        for bracket in struct.brackets:
            output += " {Range: %r-%r}" % (chr(bracket.start), chr(bracket.end))

    if getattr(struct, "chars", None):
        output += " {Chars: %r}" % struct.chars

    return output + "\n"


def disasm_regex(regex: ParsedRegex) -> str:

    output: str = format_struct_to_single_line(regex.header)

    # pretty_print_structure(regex.header)

    for instruction in regex.instructions:
        output += format_struct_to_single_line(instruction)

    return output


if __name__ == "__main__":
    SAMPLE_REGEXES: List[str] = [
        b"\x02\x00\x01\x00\x03\x04\x13\x00\x00\x11\x10\x00\x00\x00\x00A\x00\x00\x00A\x00\x00\x00C\x00\x00\x00C\x00\x00\x00H\x00\x00\x00H\x00\x00\x00L\x00\x00\x00M\x00\x00\x00Q\x00\x00\x00Q\x00\x00\x00S\x00\x00\x00T\x00\x00\x00V\x00\x00\x00V\x00\x00\x00Z\x00\x00\x00Z\x00\x00\x00a\x00\x00\x00a\x00\x00\x00c\x00\x00\x00c\x00\x00\x00h\x00\x00\x00h\x00\x00\x00l\x00\x00\x00m\x00\x00\x00q\x00\x00\x00q\x00\x00\x00s\x00\x00\x00t\x00\x00\x00v\x00\x00\x00v\x00\x00\x00z\x00\x00\x00z\x00\x00\x00\x14\x00\x00\x13\x01\x00\x1c\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff\xff\xff\x01'\x01\x00\x00\x11\x10\x00\x00\x00\x01A\x00\x00\x00A\x00\x00\x00C\x00\x00\x00C\x00\x00\x00H\x00\x00\x00H\x00\x00\x00L\x00\x00\x00M\x00\x00\x00Q\x00\x00\x00Q\x00\x00\x00S\x00\x00\x00T\x00\x00\x00V\x00\x00\x00V\x00\x00\x00Z\x00\x00\x00Z\x00\x00\x00a\x00\x00\x00a\x00\x00\x00c\x00\x00\x00c\x00\x00\x00h\x00\x00\x00h\x00\x00\x00l\x00\x00\x00m\x00\x00\x00q\x00\x00\x00q\x00\x00\x00s\x00\x00\x00t\x00\x00\x00v\x00\x00\x00v\x00\x00\x00z\x00\x00\x00z\x00\x00\x00\x14\x01\x00\x00",
        b"\x00\x00\x07\x00\x03\x04\x1c\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01\x14\x00\x00\x00\x0c-\x1c\x01\x00\x00\x00\x00\x00\x00\x00\xff\xff\xff\xff\x014\x00\x00\x00\x11\x01\x00\x00\x00\x000\x00\x00\x009\x00\x00\x00\x1c\x02\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01H\x00\x00\x00\x0c.\x1c\x03\x00\x00\x00\x01\x00\x00\x00\xff\xff\xff\xff\x01h\x00\x00\x00\x11\x01\x00\x00\x00\x000\x00\x00\x009\x00\x00\x00\x18\x06\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x01\x04\xc6\x00\x00\x00\x0cE\x1c\x04\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x01\xa9\x00\x00\x00\x11\x02\x00\x00\x00\x00+\x00\x00\x00+\x00\x00\x00-\x00\x00\x00-\x00\x00\x00\x1c\x05\x00\x00\x00\x01\x00\x00\x00\xff\xff\xff\xff\x01\xc1\x00\x00\x00\x11\x00\x00\x00\x00\x02\x19h\x00\x00\x00\x00",
        b"\x02\x00\x02\x00\x00\x00\x18\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x01\x00\x01\x04(\x00\x00\x00\x13\x00\x00\n\x04new_\x14\x00\x00\x19\x00\x00\x00\x00\x13\x01\x00\x1c\x01\x00\x00\x00\x00\x00\x00\x00\xff\xff\xff\xff\x01>\x00\x00\x00\x05\x14\x01\x00\x00",
        bytes.fromhex(
            "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"
        ),
        bytes.fromhex(
            "04000b000106011c0000000000000000ffffffff01190000001100000000040b034154201300001c0100000000000000ffffffff0034000000051400001c020000000000000001000000014b0000000c200c281301000f6400000004040c460c490c4c0c45106c0100000f8c00000004040c480c540c540c501c03000000000000000100000001870000000c53106c0100000fa000000004040c420c4c0c4f0c42106c0100000fcc00000004040c430c480c520c4f0c4d0c450c2d0c450c580c540c450c4e0c530c490c4f0c4e106c0100000fe400000004040c4e0c410c540c490c560c45106c0100000ff800000004040c450c560c410c4c106c0100000f1201000004040c570c450c420c500c410c430c4b106c0100000f3401000004040c3c0c410c4e0c4f0c4e0c590c4d0c4f0c550c530c3e106c0100000f4201000004040c2f106c0100000f680100000404110200000000410000005a000000610000007a0000000c3a0c5c106c0100000c5c0c5c1c0400000000000000ffffffff007f0100000514010018060000000000000001000000020003000104be0100000c3a1302001c0500000001000000ffffffff01b6010000110000000002140200198201000018080000000000000001000000030004000104fa0100000c3a1303001c0700000001000000ffffffff01f201000011000000000214030019be0100001c090000000000000001000000010e0200000c291c0a00000000000000ffffffff01260200001100000000040200"
        ),
        bytes.fromhex(
            "0100020002040f1c00000004040a0e6861734f776e50726f706572747910870000000f5c00000004041300000a0866756e6374696f6e1400001c0000000000000000ffffffff004600000005170001040100010057000000075c07280010870000000a0520666f72201c0100000001000000ffffffff007600000005170001040100010087000000075c075d0000"
        ),
        bytes.fromhex(
            "0000000002041106000000002400000024000000280000002b0000002e0000002e0000003f0000003f0000005b0000005e0000007b0000007d00000000"
        ),
        bytes.fromhex(
            "010002000006011300000f2a00000004040a04687474701c000000000000000001000000012500000007731042000000077707731c010000000000000001000000014200000007731400000a033a2f2f00"
        ),
        bytes.fromhex(
            "0000010000041c0000000001000000ffffffff01200000001101000000003d0000003d0000000200"
        ),
        bytes.fromhex("0000000002040f0f0000000604010722101200000007220200"),
        # WIP .. . )
    ]

    print()
    for regex in SAMPLE_REGEXES:
        print(disasm_regex(parse_regex(84, BytesIO(regex))))
        print(
            "==> Decompiled:", decompile_regex(parse_regex(84, BytesIO(regex)))
        )  # WIP ..
        print()
        print()  # WIP .. .

    # WIP .. Extra unit tests?
